/**
 * @file curlfall.cpp
 * @author askryabin
 *
 */

#include "curl/curl.h"

#include "event.h"
#include "sys/time.h"

#include <vector>
#include <iostream>
#include <ctime>

//for logging
char*
rtrimnl(char* str)
{
  size_t len = ::strlen(str);
  if (str[len-1] == '\n')
    str[len-1] = '\0';
  return str;
}

char* get_time()
{
  static char buf[64];
  timeval tv;
  gettimeofday(&tv, 0);

  strcpy(buf, rtrimnl(::ctime(&tv.tv_sec)));
  sprintf(buf + strlen(buf), ": %d", tv.tv_usec);
  return buf;
}

#define LOG std::cout << get_time() << ": "
using std::endl;

typedef std::vector<event*> ev_vector;
ev_vector events;

//set via curl callback
long g_timeout_ms = -1;
timeval g_timeout_origin;

char g_error[CURL_ERROR_SIZE];

CURLM* hmulti;

//called by curl
int
timer_callback(CURLM *multi, long timeout_ms, void *userp)
{
  gettimeofday(&g_timeout_origin, 0);
  g_timeout_ms = timeout_ms;
  return 0;
}

size_t
write_header_callback(
  void*   data, size_t  blocksize, size_t  nblocks, void* userp)
{
  size_t n = blocksize * nblocks;
  LOG << "Got header: " << std::string(static_cast<char*>(data), n) << endl;
  return n;
}

size_t
write_body_callback(
  void*   data, size_t  blocksize, size_t  nblocks, void* userp)
{
  size_t n = blocksize * nblocks;
  LOG << "Got body: " << std::string(static_cast<char*>(data), n) << endl;
  return n;
}

size_t
debug_callback(CURL* curl, curl_infotype type, char* data, size_t sz, void*)
{
  LOG << "DEBUG: info=" << type << ": " << std::string(data, sz) << endl;
  return 0;
}

void
new_transfer(CURL* heasy, const char* url)
{
  // set cURL easy handle options...
//  curl_easy_setopt(heasy, CURLOPT_NOSIGNAL, 1L);
//    curl_easy_setopt(heasy, CURLOPT_FORBID_REUSE, 1L);
//  curl_easy_setopt(heasy, CURLOPT_MAXCONNECTS, 1L);
//  curl_easy_setopt(heasy, CURLOPT_FRESH_CONNECT, 1L);
//  curl_easy_setopt(heasy, CURLOPT_TCP_NODELAY, 1L);

  //force close connection
  curl_slist* custom_headers = ::curl_slist_append(0, "Connection: Close");
  curl_easy_setopt(heasy, CURLOPT_HTTPHEADER, custom_headers);

  curl_easy_setopt(heasy, CURLOPT_USERAGENT, "curlfall");

  // enable/disable HTTP redirects
  curl_easy_setopt(heasy, CURLOPT_FOLLOWLOCATION, 0L);
  //cURL::setopt(heasy_, CURLOPT_MAXREDIRS, 10);

  // set write HTTP header callback
  curl_easy_setopt(heasy, CURLOPT_HEADERFUNCTION, &write_header_callback);

  // set write HTTP content callback
  curl_easy_setopt(heasy, CURLOPT_WRITEFUNCTION, &write_body_callback);

  // set debug HTTP header callback
  curl_easy_setopt(heasy, CURLOPT_DEBUGFUNCTION, &debug_callback);
  curl_easy_setopt(heasy, CURLOPT_VERBOSE, 1);

  curl_easy_setopt(heasy, CURLOPT_URL, url);
}

void
curl_apply()
{
  CURLMcode ecode = CURLM_OK;
  int nrunning;
  while(true)
  {
    ecode = ::curl_multi_perform(hmulti, &nrunning);
    LOG << "curl_apply: code = " << ecode << ", nrunning = " << nrunning << endl;
    if(ecode != CURLM_CALL_MULTI_PERFORM)
    {
      break;
    }
  }
}

void
fd_callback(int fd, short type, void* userp)
{
  LOG << "fd_callback: fd = " << fd << ", type = " << type << endl;

  int curl_type = 0;
  if (type & EV_READ) curl_type |= CURL_CSELECT_IN;
  if (type & EV_WRITE) curl_type |= CURL_CSELECT_OUT;

  CURLMcode ecode = CURLM_OK;
  int nrunning;
  while(true)
  {
    ecode = ::curl_multi_socket_action(hmulti, fd, curl_type, &nrunning);
    LOG << "curl_apply: code = " << ecode << ", nrunning = " << nrunning << endl;

    if(ecode == CURLM_BAD_SOCKET)
    {
      LOG << " - detected stray socket: " << fd << endl;
      //???
    }
    if(ecode != CURLM_CALL_MULTI_PERFORM)
    {
      break;
    }
  }
}

void
timeout_callback(int fd, short type, void* userp)
{
  LOG << "timeout_callback: type = " << type << endl;
  curl_apply();
}

int
socket_dummy_callback(
  CURL *easy, curl_socket_t s, int what, void *userp, void *socketp)
{
  LOG << "socket_dummy_callback: fd = " << s << ", what = " << what << endl;
  return 0;
}

void
add_curl_poll_handlers_fdset(fd_set* fs, short int type)
{
  char* raw_fdset = reinterpret_cast<char*>(fs);

  for (size_t byte_n = 0; byte_n < __NFDBITS / 8; ++byte_n)
  {
    if (raw_fdset[byte_n] == 0)
      continue;
    for (size_t bit_n = 0; bit_n < 8; ++bit_n)
    {
      if ((raw_fdset[byte_n] & (1 << bit_n)))
      {
        int fd = byte_n * 8 + bit_n;
        LOG << "Adding poll fd " << fd << ", type " << type << endl;

        event* e = new event;
        event_set(e, fd, type, &fd_callback, 0);
        event_add(e, 0);
        events.push_back(e);
      }
    }
  }
}

void
add_curl_poll_handlers()
{
  fd_set fdread;
  fd_set fdwrite;
  fd_set fdexcep;
  int maxfd;

  FD_ZERO(&fdread);
  FD_ZERO(&fdwrite);
  FD_ZERO(&fdexcep);

  ::curl_multi_fdset(hmulti, &fdread, &fdwrite, &fdexcep, &maxfd);

  add_curl_poll_handlers_fdset(&fdread, EV_READ);
  add_curl_poll_handlers_fdset(&fdwrite, EV_WRITE);
//  add_curl_poll_handlers_fdset(&fdexcep, EV_SIGNAL);
}

void
do_transfers(int ntransfers)
{
  struct event ev_timeout;
  struct timeval tv_timeout, tv_now, tv_elapsed;

  curl_apply();

  for (int i = 0; ntransfers /*&& i < 100*/; ++i)
  {
    event_set(&ev_timeout, -1, 0, &timeout_callback, 0);

    gettimeofday(&tv_now, 0);
    evutil_timerclear(&tv_timeout);

    tv_timeout.tv_sec = g_timeout_ms / 1000; //ms to sec
    tv_timeout.tv_usec = (g_timeout_ms % 1000) * 1000; //ms to us

    timeradd(&g_timeout_origin, &tv_timeout, &tv_timeout);
    timersub(&tv_timeout, &tv_now, &tv_timeout);

    event_add(&ev_timeout, &tv_timeout);

    long timeout_ms = tv_timeout.tv_sec*1000 + tv_timeout.tv_usec/1000;
    LOG << "loop " << i << " (transfers: " << ntransfers <<
      "), poll timeout " << timeout_ms << " ms..." << endl;

    add_curl_poll_handlers();
    event_loop(EVLOOP_ONCE);

    for (ev_vector::iterator it = events.begin(); it != events.end(); ++it)
    {
      event_del(*it);
      delete *it;
    }
    events.clear();
    event_del(&ev_timeout);

    //finished transfers?
    do {
      int msgleft = 0;
      CURLMsg* info = ::curl_multi_info_read(hmulti, &msgleft);
      if (!info) break;
      if (info->msg == CURLMSG_DONE)
      {
        --ntransfers;
      }
    } while (true);

    ::usleep(10000); //10ms, to reduce log output
  }
}

void
transfer(CURL* easy, const char* url)
{
  easy = ::curl_easy_init();
  LOG << "transfer of " << url << ": easy = " << easy << endl;

  new_transfer(easy, url);
  ::curl_multi_add_handle(hmulti, easy);

  do_transfers(1);

  ::curl_multi_remove_handle(hmulti, easy);
  ::curl_easy_cleanup(easy);
}

int main(int argc, char** argv)
{
  if (argc != 2)
  {
    std::cerr << "Usage: curlfall url" << endl;
    return 1;
  }
  const char* url = argv[1];

  ::curl_global_init(CURL_GLOBAL_ALL);
  event_init();

  hmulti = ::curl_multi_init();
  curl_multi_setopt(hmulti, CURLMOPT_TIMERFUNCTION, &timer_callback);
  curl_multi_setopt(hmulti, CURLMOPT_SOCKETFUNCTION, &socket_dummy_callback);

  CURL* easy1;
  transfer(easy1, url);

  CURL* easy2;
  transfer(easy2, url);

  ::curl_global_cleanup();
  return 0;
}
